#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>
#include <elf.h>
#include <boot/elf_boot.h>
#include <ip_checksum.h>
#include <x86/x86-linux.h>
#include "kexec.h"
#include "kexec-x86.h"


int bzImage_probe(FILE *file)
{
	struct x86_linux_header header;
	if (fseek(file, 0, SEEK_SET) < 0) {
		fprintf(stderr, "seek error: %s\n",
			strerror(errno));
		return -1;
	}
	if (fread(&header, sizeof(header), 1, file)  != 1) {
		fprintf(stderr, "read error: %s\n",
			strerror(errno));
		return -1;
	}
	if (memcmp(header.header_magic, "HdrS", 4) != 0) {
		fprintf(stderr, "Not a bzImage\n");
		return -1;
	}
	if (header.boot_sector_magic != 0xAA55) {
		/* No x86 boot sector present */
		return 0;
	}
	if (header.protocol_version < 0x0200) {
		/* Must be at least protocol version 2.00 */
		return 0;
	}
	if ((header.loadflags & 1) == 0) {
		/* Not a bzImage */
		return 0;
	}
	/* I've got a bzImage */
	return 1;
}


int bzImage_load(FILE *file, char *command_line, 
	void **ret_entry, struct kexec_segment **ret_segments, int *ret_nr_segments)
{
	struct x86_linux_header header;
	struct x86_linux_param_header *real_mode;
	int setup_sects;
	char *kernel_version;
	struct kexec_segment *segment;
	int nr_segments;
	long length;
	size_t size;
	int command_line_len;
	int command_line_off;
	int kern16_size;
	int setup16_off;
	char *cmdline;
	char *start16;
	unsigned long entry;
	command_line_len = strlen(command_line) +1;

	if (fseek(file, 0, SEEK_SET) < 0) {
		fprintf(stderr, "seek error: %s\n",
			strerror(errno));
		return -1;
	}
	if (fread(&header, sizeof(header), 1, file)  != 1) {
		fprintf(stderr, "read error: %s\n",
			strerror(errno));
		return -1;
	}
	setup_sects = header.setup_sects;
	if (setup_sects == 0) {
		setup_sects = 4;
	}
	kern16_size = (setup_sects +1) *512;
	kernel_version = ((unsigned char *)&header) + 512 + header.kver_addr;
	nr_segments = 2;
	segment = malloc(nr_segments *sizeof(*segment));
	if (segment == 0) {
		fprintf(stderr, "malloc failed: %s\n",
			strerror(errno));
		return -1;
	}
	/* The x86 code segment */
	command_line_off = kern16_size;
	setup16_off = (command_line_off + command_line_len + setup16_align -1) & ~(setup16_align -1);
	size = setup16_off + setup16_size;
	segment[0].bufsz = size;
	segment[0].memsz = size;
	segment[0].mem = (void *)0x90000;
	segment[0].buf = malloc(size);
	if (segment[0].buf == 0) {
		fprintf(stderr, "malloc failed: %s\n",
			strerror(errno));
		return -1;
	}
	real_mode = segment[0].buf;
	cmdline = (char *)segment[0].buf + command_line_off;
	start16 =(char *)segment[0].buf + setup16_off;
	entry = (unsigned long)segment[0].mem + setup16_off;
	if (fseek(file, 0, SEEK_SET) < 0) {
		fprintf(stderr, "seek error: %s\n",
			strerror(errno));
		return -1;
	}
	if (fread(segment[0].buf, kern16_size, 1, file)  != 1) {
		fprintf(stderr, "read error: %s\n",
			strerror(errno));
		return -1;
	}
	/* 
	 * Initialize the param_header with bootloader information.
	 */
	/* The location of the command line */
	real_mode->cl_magic = CL_MAGIC_VALUE;
	real_mode->cl_offset = kern16_size;
	if (header.protocol_version >= 0x0202) {
		real_mode->cmd_line_ptr = 0x90000 + command_line_off;
	}
	/* The loader type */
	real_mode->loader_type = LOADER_TYPE_UNKNOWN;
	/*
	 * Initialize the 16bit start information
	 */
	setup16_regs.cs = 0x9020;
	setup16_regs.ip = 0;
	setup16_regs.ss = 0x8000;
	setup16_regs.esp = 0xFFFC;
	/*
	 * Copy it all into the startup vector
	 */
	memcpy(cmdline, command_line, command_line_len);
	memcpy(start16, setup16_start, setup16_size);

	/* The 32bit data */
	if (fseek(file, 0, SEEK_END) < 0) {
		fprintf(stderr, "seek error: %s\n",
			strerror(errno));
		return -1;
	}
	length = ftell(file);
	if (length == -1) {
		fprintf(stderr, "ftell error: %s\n",
			strerror(errno));
		return -1;
	}
	if (fseek(file, kern16_size, SEEK_SET) < 0) {
		fprintf(stderr, "seek error: %s\n",
			strerror(errno));
		return -1;
	}
	size = length - kern16_size;
	segment[1].bufsz = size;
	segment[1].memsz = size;
	segment[1].mem = (void *)0x100000;
	segment[1].buf = malloc(size);
	if (segment[1].buf == 0) {
		fprintf(stderr, "malloc failed: %s\n",
			strerror(errno));
		return -1;
	}
	if (fread(segment[1].buf, size, 1, file)  != 1) {
		fprintf(stderr, "read error: %s\n",
			strerror(errno));
		return -1;
	}
	*ret_entry = (void *)entry;
	*ret_nr_segments = nr_segments;
	*ret_segments = segment;
	return 0;
}
